@@ -23,6 +23,7 @@
 
 #include "gck-crypto.h"
 
+#include "egg/egg-libgcrypt.h"
 #include "egg/egg-secure-memory.h"
 
 /* ----------------------------------------------------------------------------
@@ -198,12 +199,15 @@ unpad_rsa_pkcs1 (guchar bt, guint n_modulus, const guchar* padded,
                  gsize n_padded, gsize *n_raw)
 { 
 	const guchar *at;
+	guint check;
 	guchar *raw;
 	
-	/* The absolute minimum size including padding */
-	g_return_val_if_fail (n_modulus / 8 >= 3 + 8, NULL);
+	check = (n_modulus + 7) / 8;
 	
-	if (n_padded != n_modulus / 8)
+	/* The absolute minimum size including padding */
+	g_return_val_if_fail (check >= 3 + 8, NULL);
+	
+	if (n_padded != check)
 		return NULL;
 		
 	/* Check the header */
@@ -898,7 +902,7 @@ gck_crypto_rsa_pad_raw (guint n_modulus, const guchar* raw,
          *   padding               data
          */
 
-	total = n_modulus / 8;
+	total = (n_modulus + 7) / 8;
 	n_pad = total - n_raw;
 	if (n_pad < 0) /* minumum padding */
 		return NULL;
@@ -923,7 +927,7 @@ gck_crypto_rsa_pad_one (guint n_modulus, const guchar* raw,
          *      type  padding              data
          */
 
-	total = n_modulus / 8;
+	total = (n_modulus + 7) / 8;
 	n_pad = total - 3 - n_raw;
 	if (n_pad < 8) /* minumum padding */
 		return NULL;
@@ -949,7 +953,7 @@ gck_crypto_rsa_pad_two (guint n_modulus, const guchar* raw,
          *      type  padding              data
          */
 
-	total = n_modulus / 8;
+	total = (n_modulus + 7) / 8;
 	n_pad = total - 3 - n_raw;
 	if (n_pad < 8) /* minumum padding */
 		return NULL;
@@ -981,90 +985,9 @@ gck_crypto_rsa_unpad_two (guint bits, const guchar *padded,
  * INITIALIZATION
  */
 
-static void
-log_handler (gpointer unused, int unknown, const gchar *msg, va_list va)
-{
-	/* TODO: Figure out additional arguments */
-	g_logv ("gcrypt", G_LOG_LEVEL_MESSAGE, msg, va);
-}
-
-static int 
-no_mem_handler (gpointer unused, size_t sz, unsigned int unknown)
-{
-	/* TODO: Figure out additional arguments */
-	g_error ("couldn't allocate %lu bytes of memory", 
-	         (unsigned long int)sz);
-	return 0;
-}
-
-static void
-fatal_handler (gpointer unused, int unknown, const gchar *msg)
-{
-	/* TODO: Figure out additional arguments */
-	g_log ("gcrypt", G_LOG_LEVEL_ERROR, "%s", msg);
-}
-
-static int
-glib_thread_mutex_init (void **lock)
-{
-	*lock = g_mutex_new ();
-	return 0;
-}
-
-static int 
-glib_thread_mutex_destroy (void **lock)
-{
-	g_mutex_free (*lock);
-	return 0;
-}
-
-static int 
-glib_thread_mutex_lock (void **lock)
-{
-	g_mutex_lock (*lock);
-	return 0;
-}
-
-static int 
-glib_thread_mutex_unlock (void **lock)
-{
-	g_mutex_unlock (*lock);
-	return 0;
-}
-
-static struct gcry_thread_cbs glib_thread_cbs = {
-	GCRY_THREAD_OPTION_USER, NULL,
-	glib_thread_mutex_init, glib_thread_mutex_destroy,
-	glib_thread_mutex_lock, glib_thread_mutex_unlock,
-	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL 
-};
 
 void
 gck_crypto_initialize (void)
 {
-	static gsize gcrypt_initialized = FALSE;
-	unsigned seed;
-
-	if (g_once_init_enter (&gcrypt_initialized)) {
-		
-		/* Only initialize libgcrypt if it hasn't already been initialized */
-		if (!gcry_control (GCRYCTL_INITIALIZATION_FINISHED_P)) {
-			gcry_control (GCRYCTL_SET_THREAD_CBS, &glib_thread_cbs);
-			gcry_check_version (LIBGCRYPT_VERSION);
-			gcry_set_log_handler (log_handler, NULL);
-			gcry_set_outofcore_handler (no_mem_handler, NULL);
-			gcry_set_fatalerror_handler (fatal_handler, NULL);
-			gcry_set_allocation_handler ((gcry_handler_alloc_t)g_malloc, 
-			                             (gcry_handler_alloc_t)egg_secure_alloc, 
-			                             egg_secure_check, 
-			                             (gcry_handler_realloc_t)egg_secure_realloc, 
-			                             egg_secure_free);
-			gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
-		}
-		
-		gcry_create_nonce (&seed, sizeof (seed));
-		srand (seed);
-		
-		g_once_init_leave (&gcrypt_initialized, 1);
-	}
+	egg_libgcrypt_initialize ();
 }
